/*
   +----------------------------------------------------------------------+
   | HipHop for PHP                                                       |
   +----------------------------------------------------------------------+
   | Copyright (c) 2010-present Facebook, Inc. (http://www.facebook.com)  |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
*/

#include "hphp/runtime/vm/runtime-compiler.h"

#include "hphp/runtime/base/request-info.h"
#include "hphp/runtime/base/static-string-table.h"
#include "hphp/runtime/base/unit-cache.h"

#include "hphp/runtime/vm/as.h"
#include "hphp/runtime/vm/native.h"
#include "hphp/runtime/vm/unit-emitter.h"
#include "hphp/runtime/vm/unit-parser.h"

#include "hphp/util/assertions.h"
#include "hphp/util/sha1.h"

#include "hphp/zend/zend-string.h"

#include <folly/Range.h>

namespace HPHP {

TRACE_SET_MOD(runtime);

namespace {

///////////////////////////////////////////////////////////////////////////////

std::unique_ptr<UnitEmitter> parse(LazyUnitContentsLoader& loader,
                                   const char* filename,
                                   const Native::FuncTable& nativeFuncs,
                                   Unit** releaseUnit,
                                   bool isSystemLib,
                                   bool forDebuggerEval) {
  if (!filename) filename = "";

  tracing::Block _{
    "parse",
    [&] {
      return tracing::Props{}
        .add("filename", filename ? filename : "")
        .add("code_size", loader.fileLength());
    }
  };
  auto const wasLoaded = loader.didLoad();
  SCOPE_EXIT {
    if (!wasLoaded && loader.didLoad()) {
      tracing::updateName("parse-load");
    }
  };

  // Do not count memory used during parsing/emitting towards OOM.
  MemoryManager::SuppressOOM so(*tl_heap);

  SCOPE_ASSERT_DETAIL("parsing file") { return filename; };
  std::unique_ptr<Unit> unit;
  SCOPE_EXIT {
    if (unit && releaseUnit) *releaseUnit = unit.release();
  };

  // We don't want to invoke the JIT when trying to run PHP code.
  auto const prevFolding = RID().getJitFolding();
  RID().setJitFolding(true);
  SCOPE_EXIT { RID().setJitFolding(prevFolding); };

  std::unique_ptr<UnitEmitter> ue;
  // Check if this file contains raw hip hop bytecode instead of
  // php.  This is dictated by a special file extension.
  if (RuntimeOption::EvalAllowHhas) {
    if (const char* dot = strrchr(filename, '.')) {
      const char hhbc_ext[] = "hhas";
      if (!strcmp(dot + 1, hhbc_ext)) {
        auto const& contents = loader.contents();
        ue = assemble_string(
          contents.data(),
          contents.size(),
          filename,
          loader.sha1(),
          nativeFuncs
        );
      }
    }
  }

  // If ue != nullptr then we assembled it above, so don't feed it into
  // the extern compiler
  if (!ue) {
    auto uc = UnitCompiler::create(
      loader,
      filename,
      nativeFuncs,
      isSystemLib,
      forDebuggerEval
    );
    assertx(uc);
    tracing::BlockNoTrace _{"unit-compiler-run"};
    SCOPE_EXIT {
      if (!wasLoaded && loader.didLoad()) {
        tracing::updateName("unit-compiler-run-load");
      }
    };
    try {
      bool ignore;
      ue = uc->compile(ignore);
    } catch (const CompilerAbort& exn) {
      fprintf(stderr, "%s", exn.what());
      _Exit(1);
    }
  }

  assertx(ue);
  return ue;
}

///////////////////////////////////////////////////////////////////////////////

}

Unit* compile_file(LazyUnitContentsLoader& loader,
                   const char* filename,
                   const Native::FuncTable& nativeFuncs,
                   Unit** releaseUnit) {
  assertx(!filename || filename[0] != '/' || filename[1] != ':');
  return parse(
    loader,
    filename,
    nativeFuncs,
    releaseUnit,
    false,
    false
  )->create().release();
}

Unit* compile_string(const char* s,
                     size_t sz,
                     const char* fname,
                     const Native::FuncTable& nativeFuncs,
                     const RepoOptions& options,
                     bool isSystemLib,
                     bool forDebuggerEval) {
  // If the file is too large it may OOM the request
  MemoryManager::SuppressOOM so{*tl_heap};

  auto const name = fname ? fname : "";
  auto const sha1 = SHA1{mangleUnitSha1(
    string_sha1(folly::StringPiece{s, sz}), name, options.flags()
  )};
  LazyUnitContentsLoader loader{sha1, {s, sz}, options.flags()};
  return parse(
    loader,
    fname,
    nativeFuncs,
    nullptr,
    isSystemLib,
    forDebuggerEval
  )->create().release();
}

Unit* compile_systemlib_string(const char* s, size_t sz, const char* fname,
                               const Native::FuncTable& nativeFuncs) {
  assertx(fname && fname[0] == '/' && fname[1] == ':');
  if (RuntimeOption::RepoAuthoritative) {
    if (auto u = lookupSyslibUnit(makeStaticString(fname), nativeFuncs)) {
      return u;
    }
  }

  auto const sha1 = SHA1{mangleUnitSha1(
    string_sha1(folly::StringPiece{s, sz}),
    fname,
    RepoOptions::defaults().flags()
  )};
  LazyUnitContentsLoader loader{sha1, {s, sz}, RepoOptions::defaults().flags()};
  auto ue = parse(
    loader,
    fname,
    nativeFuncs,
    nullptr,
    true,
    false
  );
  always_assert(ue);

  auto u = ue->create();
  SystemLib::registerUnitEmitter(std::move(ue));
  return u.release();
}

Unit* compile_debugger_string(
  const char* s, size_t sz, const RepoOptions& options
) {
  return compile_string(
    s,
    sz,
    nullptr,
    Native::s_noNativeFuncs,
    options,
    false,
    true
  );
}

}
